import fileinput
import glob
import gzip
import hashlib
import re
import shutil
import urllib.request
from pathlib import Path
from typing import IO, Any, Dict, Set, Union

FASTQ_REGEX = re.compile(r"\.f(ast)?q(\.gz)?$")
PathLike = Union[str, Path]


def is_fastq(p: PathLike) -> bool:
    return Path(p).is_file() and bool(FASTQ_REGEX.search(str(p)))


def find_fastq_files(src: Path, recursive: bool = False) -> Set[str]:
    return set(p for p in glob.iglob(f"{src}/**", recursive=recursive) if is_fastq(p))


def which_open(fname: PathLike, mode: str) -> IO:
    if str(fname).endswith(".gz"):
        return gzip.open(fname, mode=mode)
    return open(fname, mode=mode)


def concatenate_fastqs(files: Set[PathLike], dest: PathLike):
    with gzip.open(dest, compresslevel=6, mode="wb") as fout, fileinput.input(
        files, openhook=which_open, mode="rb"
    ) as fin:
        for line in fin:
            if line.rstrip():  # skip empty lines
                fout.write(line)


def parse_verbose_filter_params(filters_dict: Dict[Any, Any]) -> str:
    opts = [
        ("d", "min_depth"),
        ("D", "max_depth"),
        ("q", "min_qual"),
        ("s", "min_strand_bias"),
        ("b", "min_bqb"),
        ("m", "min_mqb"),
        ("r", "min_rpb"),
        ("V", "min_vdb"),
        ("G", "max_sgb"),
        ("K", "min_frs"),
        ("w", "min_rpbz"),
        ("W", "max_rpbz"),
        ("C", "max_scbz"),
        ("M", "min_mq"),
        ("x", "min_fed"),
    ]
    flags = []
    for op, key in opts:
        if key in filters_dict:
            flags.append(f"-{op} {filters_dict[key]}")
    return " ".join(flags)


def fastq_prefix(path: Union[str, Path]) -> str:
    fname = Path(path).name
    return FASTQ_REGEX.sub("", fname, count=1)


def download_file(url: str, filename: Path):
    urllib.request.urlretrieve(url=url, filename=filename)


def validate_sha256(file_path: Path, expected_hash: str) -> bool:
    hash_obj = hashlib.sha256()

    # process the file in (byte) chunks
    chunk_size = 1_024 * 10_000
    with open(file_path, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            hash_obj.update(chunk)

    # get the hexadecimal representation of the sha256 hash
    actual_hash = hash_obj.hexdigest()

    return actual_hash == expected_hash


def decompress_file(
    compressed_file: Path, decompressed_file: Path, remove_compressed: bool = False
):
    """Decompress a gzip-compressed file
    `remove_compressed` indicates whether to delete the compressed file after
    successful decompression
    """
    with gzip.open(compressed_file, "rb") as f_in:
        with open(decompressed_file, "wb") as f_out:
            shutil.copyfileobj(f_in, f_out)

    if remove_compressed:
        compressed_file.unlink(missing_ok=True)



"""
Module responsible for clustering tbpore process consensus sequences.
Totally based on https://github.com/mbhall88/head_to_head_pipeline/blob/bcbc84971342a26cd0a9f0ad8df4f01dcf35c01c/analysis/transmission_clustering/eda/clustering.ipynb
"""

from itertools import chain
from pathlib import Path
from typing import Generator, List, Set

import networkx as nx
import numpy as np
import pandas as pd

DELIM = ","
PAIR_IDX = ("sample1", "sample2")


class AsymmetrixMatrixError(Exception):
    pass


def load_matrix(fpath, delim: str = DELIM, name: str = "") -> pd.Series:
    matrix = []
    with open(fpath) as instream:
        header = next(instream).rstrip()
        names = np.array(header.split(delim)[1:])
        idx = np.argsort(names)
        sorted_names = names[idx]
        for row in map(str.rstrip, instream):
            # sort row according to the name sorting
            sorted_row = np.array(row.split(delim)[1:], dtype=int)[idx]
            matrix.append(sorted_row)
    sorted_matrix = np.array(matrix)[idx]
    n_samples = len(sorted_names)
    diagonal_is_zero = all(sorted_matrix[i, i] == 0 for i in range(n_samples))
    if not diagonal_is_zero:
        raise AsymmetrixMatrixError("Distance matrix diagonal is not all zero")

    matrix_is_symmetric = np.allclose(sorted_matrix, sorted_matrix.T)
    if not matrix_is_symmetric:
        raise AsymmetrixMatrixError("Distance matrix is not symmetric")

    mx = pd.DataFrame(sorted_matrix, columns=sorted_names, index=sorted_names)
    # remove the lower triangle of the matrix and the middle diagonal
    mx = mx.where(np.triu(np.ones(mx.shape), k=1).astype(bool))
    mx = mx.stack().rename(name).astype(int)
    mx = mx.rename_axis(PAIR_IDX)

    return mx


def matrix_to_graph(
    mx: pd.Series, threshold: int, include_singletons: bool = False
) -> nx.Graph:
    edges = [(s1, s2, dist) for (s1, s2), dist in mx.items() if dist <= threshold]
    graph = nx.Graph()
    graph.add_weighted_edges_from(edges)
    if include_singletons:
        samples = set()
        for u in chain.from_iterable(mx.index):
            if u not in samples:
                graph.add_node(u)
                samples.add(u)
    return graph


def sort_clusters(clusters: Generator[Set[str], None, None]) -> List[List[str]]:
    # gets a list of sorted clusters
    clusters = [sorted(list(cluster)) for cluster in clusters]

    # sort by size and then IDs (in case of draws)
    clusters = sorted(clusters, key=lambda cluster: (-len(cluster), str(cluster)))

    return clusters


def get_clusters(psdm_matrix: Path, clustering_threshold: int) -> List[List[str]]:
    ont_mtx = load_matrix(psdm_matrix, name="nanopore")
    ont_graph = matrix_to_graph(
        ont_mtx, threshold=clustering_threshold, include_singletons=True
    )

    clusters = nx.connected_components(ont_graph)

    # sort clusters to guarantee determinism
    clusters = sort_clusters(clusters)
    return clusters


def get_formatted_clusters(clusters: List[List[str]]) -> str:
    clusters_as_strs = []
    for cluster_index, cluster in enumerate(clusters):
        cluster_as_str = "\t".join(cluster)
        cluster_description = f"Cluster #{cluster_index+1}:\t{cluster_as_str}"
        clusters_as_strs.append(cluster_description)
    return "\n".join(clusters_as_strs)


def produce_clusters(psdm_matrix: Path, threshold: int, outdir: Path) -> None:
    clusters = get_clusters(psdm_matrix, threshold)
    clusters_file = outdir / "clusters.txt"
    with open(clusters_file, "w") as clusters_fh:
        print(get_formatted_clusters(clusters), file=clusters_fh)



from pathlib import Path

TMP_NAME = ".tbpore"
repo_root = Path(__file__).parent.parent.resolve()
H37RV_genome = repo_root / "data/H37RV_genome/h37rv.fa.gz"
H37RV_mask = repo_root / "data/H37RV_genome/compass-mask.bed"
DECONTAMINATION_DB_INDEX = (
    Path.home() / ".tbpore/decontamination_db/remove_contam.map-ont.mmi"
)
DECONTAMINATION_DB_METADATA = repo_root / "data/decontamination_db/remove_contam.tsv.gz"
CACHE_DIR = Path.home() / ".cache"
CONFIG_FILE = repo_root / ".config.yaml"
EXTERNAL_SCRIPTS_DIR = repo_root / "external_scripts"



import logging
from enum import Enum
from pathlib import Path
from typing import Set

import click
import pandas as pd
import pysam

IS_CONTAM_COL = "is_contaminant"


class RequiredIf(click.Option):
    def __init__(self, *args, **kwargs):
        self.required_if = kwargs.pop("required_if")
        assert self.required_if, "'required_if' parameter required"
        kwargs["help"] = (
            kwargs.get("help", "")
            + " NOTE: This argument is mutually inclusive with %s" % self.required_if
        ).strip()
        super(RequiredIf, self).__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        we_are_present = self.name in opts
        other_present = self.required_if in opts

        if not other_present:
            if we_are_present:
                raise click.UsageError(
                    "Illegal usage: `%s` is mutually inclusive with `%s`"
                    % (self.name, self.required_if)
                )

        return super(RequiredIf, self).handle_parse_result(ctx, opts, args)


class Classification(Enum):
    Contaminant = "contaminant"
    Unmaped = "unmapped"
    Keep = "keep"
    Other = "other"


class Classifier:
    def __init__(self, metadata_file: str):
        self.metadata = pd.read_table(
            metadata_file,
            header=None,
            names=["organism", IS_CONTAM_COL, "accession"],
            index_col="accession",
            dtype={IS_CONTAM_COL: "bool"},
        )

    def classify(self, record: pysam.AlignedSegment) -> Classification:
        if record.is_unmapped:
            return Classification.Unmaped

        ref_id = record.reference_name
        is_contam: bool = self.metadata.at[ref_id, IS_CONTAM_COL]
        if is_contam:
            return Classification.Contaminant

        return Classification.Keep


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-i",
    "--samfile",
    help="{B,CR,S}AM file of reads mapped to decontamination database.",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
)
@click.option(
    "-m",
    "--metadata",
    help=(
        "TSV file containing information about each reference in the database. Column "
        "1 is the category, column 2 is whether the reference is contamination, column "
        "3 is the accession ID for the reference."
    ),
    type=click.Path(exists=True, dir_okay=False),
    required=True,
)
@click.option(
    "-o",
    "--outdir",
    type=click.Path(dir_okay=True, writable=True),
    help=(
        "Directory to write the output files to. The files written will be named "
        "unmapped.reads, contaminant.reads, and keep.reads"
    ),
    default=".",
    show_default=True,
)
@click.option(
    "--ignore-secondary/--include-secondary",
    help="Ignore organism assignments for secondary alignments?",
    default=True,
    show_default=True,
)
@click.option("-v", "--verbose", help="Turns on debug-level logging.", is_flag=True)
def main(
    samfile: str,
    metadata: str,
    outdir: str,
    ignore_secondary: bool,
    verbose: bool,
):
    """This scripts classifies records in an alignment to a contamination database, with
    the help of a metadata file mapping reference names to whether they are
    contamination or not. It produces three files with a read identifier for each line:
      - unmapped reads
      - contaminated reads
      - reads to keep
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s]: %(message)s", level=log_level
    )
    outdir = Path(outdir)
    outdir.mkdir(exist_ok=True)

    classifier = Classifier(metadata)
    all_read_ids: Set[str] = set()
    unmapped_reads: Set[str] = set()
    contaminant_reads: Set[str] = set()
    keep_reads: Set[str] = set()

    logging.info("Classifying records in alignment file...")
    with pysam.AlignmentFile(samfile) as bam:
        for record in bam:
            read_id = record.query_name
            if read_id is None:
                logging.warning(f"Got a record with no query name\n{str(record)}")
                continue

            all_read_ids.add(read_id)
            if record.is_secondary and ignore_secondary:
                logging.debug(f"{read_id} has secondary alignment. Skipping...")
                continue

            classification = classifier.classify(record)
            if classification is Classification.Unmaped:
                unmapped_reads.add(read_id)
            elif classification is Classification.Keep:
                keep_reads.add(read_id)
            elif classification.Contaminant:
                contaminant_reads.add(read_id)
            else:
                raise NotImplementedError(
                    f"Don't know how to handle classification: {classification}"
                )

    # if any read in the pair is a "keeper" remove it from contaminants
    contaminant_reads -= keep_reads
    # reads are only unmapped if both are unmapped
    unmapped_reads -= keep_reads.union(contaminant_reads)

    assert all_read_ids == keep_reads.union(unmapped_reads, contaminant_reads)

    logging.info(f"{len(keep_reads)} reads are to be kept")
    logging.info(f"{len(contaminant_reads)} reads are contaminants")
    logging.info(f"{len(unmapped_reads)} reads are unmapped")
    logging.info("Writing output files...")

    keep_file = outdir / "keep.reads"
    keep_file.write_text("\n".join(keep_reads))
    logging.info(f"Read identifiers to keep written to {keep_file}")

    contaminant_file = outdir / "contaminant.reads"
    contaminant_file.write_text("\n".join(contaminant_reads))
    logging.info(f"Contaminant read identifiers written to {contaminant_file}")

    unmapped_file = outdir / "unmapped.reads"
    unmapped_file.write_text("\n".join(unmapped_reads))
    logging.info(f"Unmapped read identifiers written to {unmapped_file}")


if __name__ == "__main__":
    main()



# From https://github.com/mbhall88/head_to_head_pipeline/blob/master/analysis/baseline_variants/scripts/apply_filters.py

import logging
from collections import Counter
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, NamedTuple, Optional, Tuple

import click
import numpy as np
from cyvcf2 import VCF, Variant, Writer

HIST_BINS = 40


def validate_fraction(ctx, param, value):
    f = float(value)
    if f > 1.0 or f < 0.0:
        raise click.BadParameter("fractions must be between 0.0 and 1.0")
    else:
        return f


class Tags(Enum):
    Depth = "DP"
    LowDepth = "ld"
    LowFed = "lfed"
    HighDepth = "hd"
    LowQual = "lq"
    StrandBias = "sb"
    StrandDepth = "DP4"
    Pass = "PASS"
    BaseQualBias = "BQB"
    LowBaseQualBias = "lbqb"
    MapQualBias = "MQB"
    LowMapQualBias = "lmqb"
    ReadPosBias = "RPB"
    ReadPosBiasZ = "RPBZ"
    LowReadPosBias = "lrpb"
    LowReadPosBiasZ = "lrpbz"
    HighReadPosBiasZ = "hrpbz"
    SegregationBias = "SGB"
    HighSegBias = "hsgb"
    SoftClipBiasZ = "SCBZ"
    HighSoftClipBiasZ = "hscbz"
    VariantDistanceBias = "VDB"
    LowVarDistBias = "lvdb"
    LowSupport = "frs"
    LowMapQual = "lmq"
    MapQual = "MQ"

    def __str__(self) -> str:
        return str(self.value)


@dataclass
class FilterStatus:
    low_depth: bool = False
    low_fed: bool = False
    high_depth: bool = False
    low_qual: bool = False
    strand_bias: bool = False
    low_bqb: bool = False
    low_mqb: bool = False
    low_rpb: bool = False
    low_rpbz: bool = False
    high_rpbz: bool = False
    high_scbz: bool = False
    low_mapq: bool = False
    high_sgb: bool = False
    low_vdb: bool = False
    low_support: bool = False
    delim: str = ";"

    def __str__(self) -> str:
        status = []
        if self.low_depth:
            status.append(str(Tags.LowDepth))
        if self.low_fed:
            status.append(str(Tags.LowFed))
        if self.high_depth:
            status.append(str(Tags.HighDepth))
        if self.low_qual:
            status.append(str(Tags.LowQual))
        if self.strand_bias:
            status.append(str(Tags.StrandBias))
        if self.low_bqb:
            status.append(str(Tags.LowBaseQualBias))
        if self.low_mqb:
            status.append(str(Tags.LowMapQualBias))
        if self.low_mapq:
            status.append(str(Tags.LowMapQual))
        if self.low_rpb:
            status.append(str(Tags.LowReadPosBias))
        if self.low_rpbz:
            status.append(str(Tags.LowReadPosBiasZ))
        if self.high_rpbz:
            status.append(str(Tags.HighReadPosBiasZ))
        if self.high_scbz:
            status.append(str(Tags.HighSoftClipBiasZ))
        if self.high_sgb:
            status.append(str(Tags.HighSegBias))
        if self.low_vdb:
            status.append(str(Tags.LowVarDistBias))
        if self.low_support:
            status.append(str(Tags.LowSupport))

        return self.delim.join(status) if status else str(Tags.Pass)


@dataclass
class StrandDepths:
    ref_forward: int = 0
    ref_reverse: int = 0
    alt_forward: int = 0
    alt_reverse: int = 0

    @staticmethod
    def _ratio(depths: Tuple[int, int]) -> float:
        try:
            return min(depths) / sum(depths)
        except ZeroDivisionError:
            return 1.0

    @property
    def ref_depths(self) -> Tuple[int, int]:
        return self.ref_forward, self.ref_reverse

    @property
    def alt_depths(self) -> Tuple[int, int]:
        return self.alt_forward, self.alt_reverse

    @property
    def ref_ratio(self) -> float:
        return self._ratio(self.ref_depths)

    @property
    def alt_ratio(self) -> float:
        return self._ratio(self.alt_depths)

    def to_tuple(self) -> Tuple[int, int, int, int]:
        return self.ref_forward, self.ref_reverse, self.alt_forward, self.alt_reverse


class DepthTagError(Exception):
    pass


class Genotype(NamedTuple):
    allele1: int
    allele2: int

    def is_null(self) -> bool:
        """Is the genotype null. i.e. ./."""
        return self.allele1 == -1 and self.allele2 == -1

    def is_hom(self) -> bool:
        """Is the genotype homozygous"""
        if self.is_null():
            return False
        if self.allele1 == -1 or self.allele2 == -1:
            return True
        return self.allele1 == self.allele2

    def is_het(self) -> bool:
        """Is the genotype heterozyhous"""
        return not self.is_null() and not self.is_hom()

    def is_hom_ref(self) -> bool:
        """Is genotype homozygous reference?"""
        return self.is_hom() and (self.allele1 == 0 or self.allele2 == 0)

    def is_hom_alt(self) -> bool:
        """Is genotype homozygous alternate?"""
        return self.is_hom() and (self.allele1 > 0 or self.allele2 > 0)

    def alt_index(self) -> Optional[int]:
        """If the genotype is homozygous alternate, returns the 0-based index of the
        alt allele in the alternate allele array.
        """
        if not self.is_hom_alt():
            return None
        return max(self.allele1, self.allele2) - 1

    def allele_index(self) -> Optional[int]:
        """The index of the called allele"""
        if self.is_hom_ref() or self.is_null():
            return 0
        elif self.is_hom_alt():
            return self.alt_index() + 1
        else:
            raise NotImplementedError(f"Het Genotype is unexpected: {self}")

    @staticmethod
    def from_arr(arr: List[int]) -> "Genotype":
        alleles = [a for a in arr if type(a) is int]
        if len(alleles) < 2:
            alleles.append(-1)
        return Genotype(*alleles)


class Filter:
    def __init__(
        self,
        expected_depth: int = 0,
        min_depth: int = 0,
        min_fed: float = 0,
        max_depth: int = 0,
        min_strand_bias: int = 0,
        min_qual: float = 0,
        min_bqb: float = 0,
        min_mqb: float = 0,
        min_rpb: float = 0,
        max_sgb: float = 0,
        min_vdb: float = 0,
        min_frs: float = 0,
        min_mq: int = 0,
        min_rpbz: Optional[float] = None,
        max_rpbz: Optional[float] = None,
        max_scbz: Optional[float] = None,
    ):
        self.expected_depth = expected_depth
        self.min_depth = min_depth
        self.min_fed = min_fed
        self.max_depth = max_depth
        self.min_strand_bias = min_strand_bias / 100
        self.min_qual = min_qual
        self.min_bqb = min_bqb
        self.min_mqb = min_mqb
        self.min_mq = min_mq
        self.min_rpb = min_rpb
        self.max_sgb = max_sgb
        self.min_vdb = min_vdb
        self.min_frs = min_frs
        if min_rpbz is None:
            min_rpbz = -float("inf")
        self.min_rpbz = min_rpbz
        if max_rpbz is None:
            max_rpbz = float("inf")
        self.max_rpbz = max_rpbz
        if max_scbz is None:
            max_scbz or float("inf")
        self.max_scbz = max_scbz

        if self.min_depth and self.max_depth and self.min_depth > self.max_depth:
            raise ValueError(
                f"Minimum depth is more than maximum depth: "
                f"{self.min_depth:.1f} > {self.max_depth:.1f}"
            )

    def _is_low_depth(self, variant: Variant) -> bool:
        variant_depth = get_depth(variant)
        return variant_depth < self.min_depth if self.min_depth else False

    def _is_low_fed(self, variant: Variant) -> bool:
        hq_depth = get_hq_depth(variant)
        fed = hq_depth / self.expected_depth
        return fed < self.min_fed if self.min_fed > 0 else False

    def _is_high_depth(self, variant: Variant) -> bool:
        variant_depth = get_depth(variant)
        return variant_depth > self.max_depth if self.max_depth else False

    def _is_low_qual(self, variant: Variant) -> bool:
        return (variant.QUAL or 0) < self.min_qual

    def _is_low_mapq(self, variant: Variant) -> bool:
        variant_mapq = get_mapq(variant)
        return variant_mapq < self.min_mq if self.min_mq else False

    def _is_low_bqb(self, variant: Variant) -> bool:
        bqb = variant.INFO.get(str(Tags.BaseQualBias))
        return bqb is not None and bqb < self.min_bqb

    def _is_low_mqb(self, variant: Variant) -> bool:
        mqb = variant.INFO.get(str(Tags.MapQualBias))
        return mqb is not None and mqb < self.min_mqb

    def _is_low_rpb(self, variant: Variant) -> bool:
        rpb = variant.INFO.get(str(Tags.ReadPosBias))
        return rpb is not None and rpb < self.min_rpb

    def _is_high_sgb(self, variant: Variant) -> bool:
        sgb = variant.INFO.get(str(Tags.SegregationBias))
        return sgb is not None and sgb > self.max_sgb

    def _is_low_vdb(self, variant: Variant) -> bool:
        vdb = variant.INFO.get(str(Tags.VariantDistanceBias))
        return vdb is not None and vdb < self.min_vdb

    def _is_low_support(self, variant: Variant) -> bool:
        frs = fraction_read_support(variant)
        return frs < self.min_frs if self.min_frs else False

    def _is_low_rpbz(self, variant: Variant) -> bool:
        rpbz = variant.INFO.get(str(Tags.ReadPosBiasZ))
        return rpbz is not None and rpbz < self.min_rpbz

    def _is_high_rpbz(self, variant: Variant) -> bool:
        rpbz = variant.INFO.get(str(Tags.ReadPosBiasZ))
        return rpbz is not None and rpbz > self.max_rpbz

    def _is_high_scbz(self, variant: Variant) -> bool:
        scbz = variant.INFO.get(str(Tags.SoftClipBiasZ))
        return scbz is not None and scbz > self.max_scbz

    def filter_status(self, variant: Variant) -> str:
        status = FilterStatus()
        if self.min_depth or self.max_depth:
            status.low_depth = self._is_low_depth(variant)
            status.high_depth = self._is_high_depth(variant)

        if self.min_fed > 0:
            status.low_fed = self._is_low_fed(variant)

        if self.min_mq:
            status.low_mapq = self._is_low_mapq(variant)

        if self.min_qual:
            status.low_qual = self._is_low_qual(variant)

        if self.min_strand_bias:
            strand_depths = get_strand_depths(variant)
            assert strand_depths is not None, (
                f"Strand bias filter should be turned off if no {Tags.StrandDepth} "
                f"tag is present."
            )

            gt = Genotype.from_arr(variant.genotypes[0])
            if gt.is_hom_alt():
                ratio = strand_depths.alt_ratio
            elif gt.is_hom_ref():
                ratio = strand_depths.ref_ratio
            elif gt.is_het():
                ratio = min(strand_depths.ref_ratio, strand_depths.alt_ratio)
            elif gt.is_null():
                ratio = float("inf")
            else:
                raise NotImplementedError(
                    f"Don't know how to interpret genotype {gt} for variant at "
                    f"POS {variant.POS}"
                )
            status.strand_bias = ratio < self.min_strand_bias

        if self.min_bqb:
            status.low_bqb = self._is_low_bqb(variant)

        if self.min_mqb:
            status.low_mqb = self._is_low_mqb(variant)

        if self.min_rpb:
            status.low_rpb = self._is_low_rpb(variant)

        if self.min_rpbz:
            status.low_rpbz = self._is_low_rpbz(variant)

        if self.max_rpbz:
            status.high_rpbz = self._is_high_rpbz(variant)

        if self.max_scbz:
            status.high_scbz = self._is_high_scbz(variant)

        if self.max_sgb != 0:
            status.high_sgb = self._is_high_sgb(variant)

        if self.min_vdb:
            status.low_vdb = self._is_low_vdb(variant)

        if self.min_frs:
            status.low_support = self._is_low_support(variant)

        return str(status)

    def add_filters_to_header(self, vcf: VCF):
        if self.min_depth > 0:
            header = {
                "ID": str(Tags.LowDepth),
                "Description": (
                    f"Depth ({Tags.Depth}) less than {self.min_depth} - i.e., {Tags.Depth}<{self.min_depth:.1f}"
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. depth: {header}")

        if self.min_fed > 0:
            header = {
                "ID": str(Tags.LowFed),
                "Description": (
                    f"High-quality depth of the called allele as a fraction of the expected (median; {self.expected_depth}) is "
                    f"less than {self.min_fed}"
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. FED: {header}")

        if self.min_mq > 0:
            header = {
                "ID": str(Tags.LowMapQual),
                "Description": (
                    f"Mapping quality ({Tags.MapQual.value}) less than {self.min_mq} - i.e., {Tags.MapQual.value}<{self.min_mq:.0f}"
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. depth: {header}")

        if self.max_depth > 0:
            header = {
                "ID": str(Tags.HighDepth),
                "Description": (
                    f"Depth ({Tags.Depth}) more than {self.max_depth} - i.e., {Tags.Depth}>{self.max_depth:.1f}"
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for max. depth: {header}")

        if self.min_qual > 0:
            header = {
                "ID": str(Tags.LowQual),
                "Description": f"QUAL less than {self.min_qual}",
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. QUAL: {header}")

        if self.min_strand_bias > 0:
            header = {
                "ID": str(Tags.StrandBias),
                "Description": (
                    f"A strand on the called allele has less than  "
                    f"{self.min_strand_bias:.2%} of the high-quality depth for that "
                    f"allele. This is judged on the {Tags.StrandDepth} tag."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for strand bias: {header}")

        if self.min_frs > 0:
            header = {
                "ID": str(Tags.LowSupport),
                "Description": f"Fraction of read support on called allele is less than {self.min_frs}",
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. FRS: {header}")

        if self.min_bqb > 0:
            header = {
                "ID": str(Tags.LowBaseQualBias),
                "Description": (
                    f"Base Quality Bias ({Tags.BaseQualBias}) is less than "
                    f"{self.min_bqb}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. base quality bias: {header}")

        if self.min_mqb > 0:
            header = {
                "ID": str(Tags.LowMapQualBias),
                "Description": (
                    f"Mapping Quality Bias ({Tags.MapQualBias}) is less than "
                    f"{self.min_mqb}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. mapping quality bias: {header}")

        if self.min_rpb > 0:
            header = {
                "ID": str(Tags.LowReadPosBias),
                "Description": (
                    f"Read Position Bias ({Tags.ReadPosBias}) is less than "
                    f"{self.min_rpb}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. read position bias: {header}")

        if self.min_rpbz is not None:
            header = {
                "ID": str(Tags.LowReadPosBiasZ),
                "Description": (
                    f"Read Position Bias z-test score ({Tags.ReadPosBiasZ}) is less than "
                    f"{self.min_rpbz}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. read position bias z-test: {header}")

        if self.max_rpbz is not None:
            header = {
                "ID": str(Tags.HighReadPosBiasZ),
                "Description": (
                    f"Read Position Bias z-test score ({Tags.ReadPosBiasZ}) is more than "
                    f"{self.max_rpbz}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for max. read position bias z-test: {header}")

        if self.max_scbz is not None:
            header = {
                "ID": str(Tags.HighSoftClipBiasZ),
                "Description": (
                    f"Soft-Clip Length Bias z-test score ({Tags.SoftClipBiasZ}) is more than "
                    f"{self.max_scbz}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for max. soft-clip length bias z-test: {header}")

        if self.max_sgb != 0:
            header = {
                "ID": str(Tags.HighSegBias),
                "Description": (
                    f"Segregation-based metric ({Tags.SegregationBias}) is greater "
                    f"than {self.max_sgb}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for max. segregation bias: {header}")

        if self.min_vdb > 0:
            header = {
                "ID": str(Tags.LowVarDistBias),
                "Description": (
                    f"Variant distance bias ({Tags.VariantDistanceBias}) is less "
                    f"than {self.min_vdb}."
                ),
            }
            vcf.add_filter_to_header(header)
            logging.debug(f"Header for min. variant distance bias: {header}")


def fraction_read_support(variant: Variant, sample_idx: int = 0) -> float:
    strand_depths = get_strand_depths(variant)
    if strand_depths is None:
        return 1.0
    total_covg = sum(strand_depths.to_tuple())
    called_idx = Genotype.from_arr(variant.genotypes[0]).allele_index()
    if called_idx == 0:  # ref
        called_covg = sum(strand_depths.ref_depths)
    else:
        called_covg = sum(strand_depths.alt_depths)
    try:
        return called_covg / total_covg
    except ZeroDivisionError:
        return 1.0


def get_depth(variant: Variant, default: int = 0) -> int:
    return variant.INFO.get(str(Tags.Depth), default)


def get_hq_depth(variant: Variant, default: int = 0) -> int:
    strand_depths = get_strand_depths(variant)
    if strand_depths is None:
        return default
    called_idx = Genotype.from_arr(variant.genotypes[0]).allele_index()
    if called_idx == 0:  # ref
        called_covg = sum(strand_depths.ref_depths)
    else:
        called_covg = sum(strand_depths.alt_depths)
    return called_covg


def get_mapq(variant: Variant, default: int = 0) -> int:
    mq = variant.INFO.get(str(Tags.MapQual), default)
    if mq is None:  # even though we gave a default it can still return none
        mq = default
    return mq


def get_strand_depths(
    variant: Variant, default: Optional[StrandDepths] = None
) -> Optional[StrandDepths]:
    strand_depths = variant.INFO.get(str(Tags.StrandDepth), None)
    return StrandDepths(*strand_depths) if strand_depths is not None else default


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-i",
    "--in-vcf",
    help="VCF file to apply filters to.",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    default="-",
    show_default=True,
)
@click.option(
    "-o",
    "--out-vcf",
    help="New VCF with filters.",
    type=click.Path(exists=False, dir_okay=False, allow_dash=True),
    default="-",
    show_default=True,
)
@click.option(
    "-d",
    "--min-depth",
    help=(
        "Minimum read depth. This filter "
        f"has ID: {Tags.LowDepth.value}. Set to 0 to disable"
    ),
    default=0,
    show_default=True,
)
@click.option(
    "-x",
    "--min-fed",
    help=f"Minimum fraction of expected depth. Where expected depth is the median. "
    f"This filter has ID: {Tags.LowFed.value}. Set to 0 to disable",
    default=0.0,
    show_default=True,
    type=float,
)
@click.option(
    "-D",
    "--max-depth",
    help=(
        "Maximum depth. This filter "
        f"has ID: {Tags.HighDepth.value}. Set to 0 to disable"
    ),
    default=0,
    show_default=True,
)
@click.option(
    "-s",
    "--min-strand-bias",
    help=(
        "Filter a variant if either strand has less than INT% of the (high-quality) "
        f"depth on the called allele ({Tags.StrandDepth.value}). This filter has ID: "
        f"{Tags.StrandBias}. Set to 0 to disable"
    ),
    type=click.IntRange(0, 50),
    metavar="INT",
    default=25,
    show_default=True,
)
@click.option(
    "-q",
    "--min-qual",
    help=(
        f"Filter a variant if QUAL is less than INT. This filter has ID: "
        f"{Tags.LowQual.value}. Set to 0 to disable"
    ),
    default=0.0,
    show_default=True,
)
@click.option(
    "-K",
    "--min-frs",
    help=(
        f"Minimum fraction of reads supporting the called allele. "
        f"This filter has ID: {Tags.LowSupport.value}. Set to 0 to disable"
    ),
    default=0.0,
    type=float,
    show_default=True,
    callback=validate_fraction,
)
@click.option(
    "-M",
    "--min-mq",
    help=(
        f"Minimum mapping quality ({Tags.MapQual.value}) score. "
        f"This filter has ID: {Tags.LowMapQual.value}. Set to 0 to disable"
    ),
    default=0,
    type=int,
    show_default=True,
)
@click.option(
    "-b",
    "--min-bqb",
    help=(
        f"Filter a variant if base quality bias ({Tags.BaseQualBias.value}) is less "
        f"than FLOAT. This filter has ID: {Tags.LowBaseQualBias.value}. Set to 0 to "
        f"disable"
    ),
    default=0.0,
    show_default=True,
)
@click.option(
    "-m",
    "--min-mqb",
    help=(
        f"Filter a variant if mapping quality bias ({Tags.MapQualBias.value}) is "
        f"less than FLOAT. This filter has ID: {Tags.LowMapQualBias.value}. Set to 0 to "
        f"disable"
    ),
    default=0.0,
    show_default=True,
)
@click.option(
    "-r",
    "--min-rpb",
    help=(
        f"Filter a variant if read position bias ({Tags.ReadPosBias.value}) is "
        f"less than FLOAT. This filter has ID: {Tags.LowReadPosBias.value}. Set to 0 to "
        f"disable"
    ),
    default=0.0,
    show_default=True,
)
@click.option(
    "-w",
    "--min-rpbz",
    help=(
        f"Filter a variant if read position bias z-test score ({Tags.ReadPosBiasZ.value}) is "
        f"less than FLOAT. This filter has ID: {Tags.LowReadPosBiasZ.value}."
    ),
    default=None,
    type=float,
)
@click.option(
    "-W",
    "--max-rpbz",
    help=(
        f"Filter a variant if read position bias z-test score ({Tags.ReadPosBiasZ.value}) is "
        f"more than FLOAT. This filter has ID: {Tags.HighReadPosBiasZ.value}."
    ),
    default=None,
    type=float,
)
@click.option(
    "-C",
    "--max-scbz",
    help=(
        f"Filter a variant if soft-clip length bias z-test score ({Tags.SoftClipBiasZ.value}) is "
        f"more than FLOAT. This filter has ID: {Tags.HighSoftClipBiasZ.value}."
    ),
    default=None,
    type=float,
)
@click.option(
    "-G",
    "--max-sgb",
    help=(
        f"Filter a variant if segregation bias ({Tags.SegregationBias.value}) is "
        f"greater than FLOAT. This filter has ID: {Tags.HighSegBias.value}. Set to 0 to "
        f"disable"
    ),
    default=0.0,
    show_default=True,
)
@click.option(
    "-V",
    "--min-vdb",
    help=(
        f"Filter a variant if variant distance bias ({Tags.VariantDistanceBias.value}) "
        f"is less than FLOAT. This filter has ID: {Tags.LowVarDistBias.value}. Set to 0 "
        f"to disable"
    ),
    default=0.0,
    show_default=True,
)
@click.option(
    "--overwrite/--no-overwrite",
    "-f/-F",
    default=True,
    show_default=True,
    help="Overwrite existing information in FILTER field.",
)
@click.option(
    "-p/-P",
    "--hist/--no-hist",
    default=True,
    show_default=True,
    help="Print histograms of depth and QUAL. Histograms will be printed to stderr.",
)
@click.option("-v", "--verbose", help="Turns on debug-level logging.", is_flag=True)
def main(
    in_vcf: str,
    out_vcf: str,
    overwrite: bool,
    verbose: bool,
    min_qual: float,
    min_depth: int,
    min_fed: float,
    max_depth: int,
    min_strand_bias: int,
    min_bqb: float,
    min_mqb: float,
    min_rpb: float,
    min_rpbz: Optional[float],
    max_rpbz: Optional[float],
    max_scbz: Optional[float],
    max_sgb: float,
    min_vdb: float,
    hist: bool,
    min_frs: float,
    min_mq: int,
):
    """Apply the following filters to a VCF:\n
    - Minimum proportion of the expected (median) depth\n
    - Maximum proportion of the expected (median) depth\n
    - Minimum QUAL threshold\n
    - Minimum Strand bias percentage
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s]: %(message)s", level=log_level
    )

    vcf_reader = VCF(in_vcf)
    if not vcf_reader.contains(Tags.Depth.value):
        raise DepthTagError(f"Depth tag {Tags.Depth} not found in header")

    if (not vcf_reader.contains(str(Tags.StrandDepth))) and min_strand_bias:
        logging.warning(
            f"Strand depth tag {Tags.StrandDepth} not found in header. "
            f"Turning off strand bias filter..."
        )
        min_strand_bias = 0

    logging.info("Calculating expected (median) depth...")
    depths = []
    quals = []
    for v in vcf_reader:
        depths.append(get_depth(v))
        quals.append(v.QUAL or 0)

    median_depth = np.median(depths)
    logging.info(f"Expected depth: {median_depth}")

    if hist:
        import histoprint

        tick_format = "% .1f"
        logging.info("Depth histogram:")
        histoprint.print_hist(
            np.histogram(depths, bins=HIST_BINS),
            title="Depth histogram",
            summary=True,
            tick_format=tick_format,
            file=click.get_text_stream("stderr"),
        )

        logging.info("QUAL histogram")
        histoprint.print_hist(
            np.histogram(quals, bins=HIST_BINS),
            title="QUAL histogram",
            summary=True,
            tick_format=tick_format,
            file=click.get_text_stream("stderr"),
        )

    assessor = Filter(
        expected_depth=int(median_depth),
        min_qual=min_qual,
        min_depth=min_depth,
        min_fed=min_fed,
        max_depth=max_depth,
        min_strand_bias=min_strand_bias,
        min_bqb=min_bqb,
        min_mqb=min_mqb,
        min_rpb=min_rpb,
        max_sgb=max_sgb,
        min_vdb=min_vdb,
        min_frs=min_frs,
        min_mq=min_mq,
        min_rpbz=min_rpbz,
        max_rpbz=max_rpbz,
        max_scbz=max_scbz,
    )

    vcf_reader = VCF(in_vcf)
    assessor.add_filters_to_header(vcf_reader)

    if not Path(out_vcf).parent.exists():
        Path(out_vcf).parent.mkdir(exist_ok=True, parents=True)

    vcf_writer = Writer(out_vcf, tmpl=vcf_reader)

    stats = Counter()
    logging.info("Filtering variants...")
    for variant in vcf_reader:
        filter_status = assessor.filter_status(variant)

        if (
            (not overwrite)
            and variant.FILTER is not None
            and filter_status != str(Tags.Pass)
        ):
            current_filter = variant.FILTER.rstrip(";")
            variant.FILTER = f"{current_filter};{filter_status}"
        else:
            variant.FILTER = filter_status

        vcf_writer.write_record(variant)

        stats.update(filter_status.split(";"))

    vcf_reader.close()
    vcf_writer.close()

    logging.info("FILTER STATISTICS")
    logging.info("=================")
    for filt, count in stats.items():
        logging.info(f"Filter: {filt}\tCount: {count}")

    logging.info("Done!")


if __name__ == "__main__":
    main()



# from https://raw.githubusercontent.com/mbhall88/head_to_head_pipeline/master/analysis/baseline_variants/scripts/consensus.py
# with a small modification to read gzipped references

import gzip
import logging
from collections import defaultdict
from enum import Enum
from pathlib import Path
from typing import Dict, List, NamedTuple, Optional, Set, TextIO, Union

import click
from cyvcf2 import VCF, Variant

N = "N"
PathLike = Union[str, Path]
DEFAULT_HET_CHOICES = ["none", "ref", "alt"]


class UnknownDefaultHet(Exception):
    pass


class Bed:
    """Positions in this class are 1-based - in contrast to BED positions being 0-based.
    This is to align it with VCF positions which are 1-based."""

    def __init__(self, file: Optional[PathLike] = None, zero_based: bool = True):
        self.file = Path(file) if file is not None else file
        self.zero_based = zero_based
        self.positions: Dict[str, Set[int]] = defaultdict(set)
        if self.file is not None:
            with self.file.open() as instream:
                for line in map(str.rstrip, instream):
                    # start is 0-based inclusive; end is 0-based non-inclusive
                    chrom, start, end = line.split()[:3]
                    start = int(start)
                    end = int(end)
                    self.positions[chrom].update(range(start, end))

    def __contains__(self, variant: Variant) -> bool:
        chrom = variant.CHROM
        pos = variant.POS if not self.zero_based else variant.POS - 1
        return chrom in self.positions and pos in self.positions[chrom]


class Genotype(NamedTuple):
    allele1: int
    allele2: int

    def is_null(self) -> bool:
        """Is the genotype null. i.e. ./."""
        return self.allele1 == -1 and self.allele2 == -1

    def is_hom(self) -> bool:
        """Is the genotype homozygous"""
        if self.is_null():
            return False
        if self.allele1 == -1 or self.allele2 == -1:
            return True
        return self.allele1 == self.allele2

    def is_het(self) -> bool:
        """Is the genotype heterozyhous"""
        return not self.is_null() and not self.is_hom()

    def is_hom_ref(self) -> bool:
        """Is genotype homozygous reference?"""
        return self.is_hom() and (self.allele1 == 0 or self.allele2 == 0)

    def is_hom_alt(self) -> bool:
        """Is genotype homozygous alternate?"""
        return self.is_hom() and (self.allele1 > 0 or self.allele2 > 0)

    def alt_index(self) -> Optional[int]:
        """If the genotype is homozygous alternate, returns the 0-based index of the
        alt allele in the alternate allele array.
        """
        if not self.is_hom_alt():
            return None
        return max(self.allele1, self.allele2) - 1

    @staticmethod
    def from_arr(arr: List[int]) -> "Genotype":
        alleles = [a for a in arr if type(a) is int]
        if len(alleles) < 2:
            alleles.append(-1)
        return Genotype(*alleles)


class Classification(Enum):
    Ref = "REF"
    Alt = "ALT"
    Null = "NULL"
    Het = "HET"

    def __str__(self):
        return self.value

    @staticmethod
    def from_variant(variant: Variant) -> "Classification":
        gt = Genotype.from_arr(variant.genotypes[0])
        if gt.is_null():
            classification = Classification.Null
        elif gt.is_het():
            classification = Classification.Het
        elif gt.is_hom_ref():
            classification = Classification.Ref
        elif gt.is_hom_alt():
            classification = Classification.Alt
        else:
            raise NotImplementedError(
                f"Can't determine genotype for variant \n{str(variant)}"
            )
        return classification


class Classifier:
    def __init__(
        self,
        mask: Optional[Bed] = None,
        ignore_filter: bool = False,
        ignore_mask: bool = False,
        ignore_null: bool = False,
        het_default: str = "none",
    ):
        self.ignore_filter = ignore_filter
        self.ignore_mask = ignore_mask
        self.ignore_null = ignore_null
        if het_default not in DEFAULT_HET_CHOICES:
            raise UnknownDefaultHet(
                f"{het_default} is not in the known het default choices: "
                f"{DEFAULT_HET_CHOICES}"
            )
        self.het_default = het_default

        if mask is None:
            mask = Bed()
        self.mask = mask

    def classify(self, variant: Variant) -> str:
        length = len(variant.REF)
        if self.ignore_mask and variant in self.mask:
            return N * length

        if self.ignore_filter and variant.FILTER is not None:
            return N * length

        classification = Classification.from_variant(variant)
        if classification is Classification.Null:
            return N * length if self.ignore_null else variant.REF
        if classification is Classification.Het:
            if self.het_default == "none":
                return N * length
            else:
                return_ref = self.het_default == "ref"
                if return_ref:
                    return variant.REF
                ref_in_called_alleles = 0 in variant.genotypes[0]
                if not ref_in_called_alleles:
                    logging.warning(
                        "Got a het with two different ALT calls. You have specified "
                        "ALT as the default base for hets. Using ALT with highest "
                        "index..."
                    )
                return variant.ALT[max(variant.genotypes[0]) - 1]

        if classification is Classification.Ref:
            return variant.REF
        if classification is Classification.Alt:
            gt = Genotype.from_arr(variant.genotypes[0])
            idx = gt.alt_index()
            assert idx is not None
            return variant.ALT[idx]
        raise NotImplementedError(f"Unable to classify variant: {str(variant)}")


def load_reference(path: str) -> Dict[str, List[str]]:
    index = dict()
    with gzip.open(path, "rt") as instream:
        seqid = None
        for line in map(str.rstrip, instream):
            if line.startswith(">"):
                seqid = line.split()[0][1:]
                if seqid in index:
                    raise ReferenceError(
                        f"Found duplicate sequence identifier in the reference fasta: "
                        f"{seqid}"
                    )
                index[seqid] = []
                continue
            if seqid is None:
                raise ReferenceError(
                    "Did not find header before sequence in reference fasta"
                )
            index[seqid].extend(list(line))

    return index


@click.command()
@click.help_option("--help", "-h")
@click.option(
    "-i",
    "--vcf",
    help="VCF file generate the consensus for.",
    type=click.Path(exists=True, dir_okay=False, allow_dash=True),
    default="-",
    show_default=True,
)
@click.option(
    "-f",
    "--ref",
    help="Reference fasta file.",
    type=click.Path(exists=True, dir_okay=False, allow_dash=False),
    required=True,
)
@click.option(
    "-m",
    "--mask",
    "bedfile",
    help="BED file containing positions to mask.",
    type=click.Path(exists=True, dir_okay=False),
)
@click.option(
    "-o",
    "--output",
    help="File path of consensus fasta to output.",
    type=click.File(mode="w"),
    default="-",
    show_default=True,
)
@click.option(
    "-H",
    "--het-default",
    help=(
        "Which heterozygous base should be used? If 'none', an 'N' will be used in the "
        "consensus sequence"
    ),
    type=click.Choice(DEFAULT_HET_CHOICES, case_sensitive=False),
    default=DEFAULT_HET_CHOICES[0],
    show_default=True,
)
@click.option(
    "-I",
    "--ignore",
    help=(
        "Which types of variants should be ignored (i.e. replaced with 'N')? If not "
        "ignoring filters/mask and the position is called ALT, then ALT will be used. "
        "For null/missing the REF base will be used. This option can be specified "
        "multiple times. For example, to ignore null and mask positions use: -I null "
        "-I mask"
    ),
    type=click.Choice(
        ("none", "all", "missing", "null", "filter", "mask"), case_sensitive=False
    ),
    default=["all"],
    show_default=True,
    multiple=True,
)
@click.option(
    "-s",
    "--sample-id",
    help=(
        "Sample identifier to use for the output fasta file. If not specified, the "
        "sample column name in the VCF will be used. If more than one contig is "
        "present, it will be combined with the sample ID (i.e. CONTIG|SAMPLE_ID)"
    ),
)
@click.option("-v", "--verbose", help="Turns on debug-level logging.", is_flag=True)
def main(
    vcf: str,
    ref: str,
    output: TextIO,
    verbose: bool,
    bedfile: str,
    het_default: str,
    ignore: tuple,
    sample_id: str,
):
    """This script will produce a consensus fasta file from a VCF file.
    The VCF must contain ONLY SNVs; at this stage it cannot handle any other variant
    type. It will also only use the first sample in the VCF currently.
    """
    log_level = logging.DEBUG if verbose else logging.INFO
    logging.basicConfig(
        format="%(asctime)s [%(levelname)s]: %(message)s", level=log_level
    )

    if "none" in ignore and len(ignore) > 1:
        raise click.BadOptionUsage(
            "ignore",
            f"'none' cannot be used in conjunction with other --ignore options and you "
            f"passed {ignore}",
        )

    if "mask" in ignore and bedfile is None:
        logging.warning(
            "You have asked to ignore masked positions, but no mask was provided"
        )

    if bedfile:
        logging.info("Loading mask...")
        mask = Bed(bedfile)
        logging.info(f"Loaded mask for {len(mask.positions)} chromosome(s).")
    else:
        logging.info("No mask given...")
        mask = Bed()

    logging.info("Loading reference fasta...")
    consensus = load_reference(ref)
    expected_consensus_seq_lens = {k: len(s) for k, s in consensus.items()}
    logging.info(f"Loaded {len(consensus)} sequence(s) from reference fasta")

    vcf_reader = VCF(vcf)

    if not sample_id:
        sample_id = vcf_reader.samples[0]

    multiple_contigs = len(vcf_reader.seqnames) > 1
    if multiple_contigs:
        logging.info(
            f"Sequence identifiers will be CONTIG|{sample_id} where CONTIG is the name "
            f"of the relevant contig"
        )
    else:
        logging.info(f"Sequence identifier will be {sample_id}")

    ignore_filter = any(s in ignore for s in ("filter", "all"))
    ignore_null = any(s in ignore for s in ("null", "all"))
    ignore_missing = any(s in ignore for s in ("missing", "all"))
    ignore_mask = any(s in ignore for s in ("mask", "all"))
    classifier = Classifier(
        mask=mask,
        ignore_filter=ignore_filter,
        ignore_mask=ignore_mask,
        ignore_null=ignore_null,
        het_default=het_default,
    )
    warned_chroms = set()
    positions_in_vcf = defaultdict(set)
    logging.info("Generating consensus from variants...")
    for variant in vcf_reader:
        if variant.CHROM not in consensus and variant.CHROM not in warned_chroms:
            logging.warning(
                f"{variant.CHROM} is not in the reference fasta. Skipping..."
            )
            warned_chroms.add(variant.CHROM)
            continue

        bases = classifier.classify(variant)
        start_on_consensus = variant.POS - 1
        end_on_consensus = start_on_consensus + len(bases)
        consensus[variant.CHROM][start_on_consensus:end_on_consensus] = list(bases)
        for i in range(start_on_consensus, end_on_consensus):
            positions_in_vcf[variant.CHROM].add(i)
    logging.info("Consensus from variants complete")

    if ignore_missing:
        logging.info("Updating consensus with ignored missing positions...")
        for chrom in consensus:
            all_positions = set(range(len(consensus[chrom])))
            missing_positions = all_positions - positions_in_vcf.get(chrom, set())
            for pos in missing_positions:
                consensus[chrom][pos] = N
        logging.info("Finished updating missing positions")

    logging.info("Writing output consensus fasta...")
    for chrom in consensus:
        if not multiple_contigs:
            header = f">{sample_id}"
        else:
            header = f">{chrom}|{sample_id}"
        print(header, file=output)
        consensus_seq = "".join(consensus[chrom])
        expected_length = expected_consensus_seq_lens[chrom]
        assert (
            len(consensus_seq) == expected_length
        ), f"{chrom}'s consensus sequence ({len(consensus_seq)}) is not the expected length {expected_length}"
        print(consensus_seq, file=output)

    logging.info("Done")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tbpore']

package_data = \
{'': ['*']}

install_requires = \
['click>=8.0.3,<9.0.0',
 'cyvcf2>=0.30,<0.31',
 'loguru>=0.5.3,<0.6.0',
 'networkx>=2.8,<3.0',
 'pandas>=1.4.2,<2.0.0',
 'pysam>=0.17,<0.18',
 'pyyaml>=6.0']

entry_points = \
{'console_scripts': ['tbpore = tbpore.tbpore:main']}

setup_kwargs = {
    'name': 'tbpore',
    'version': '0.4.1',
    'description': 'Mycobacterium tuberculosis genomic analysis from Nanopore sequencing data',
    'long_description': '# TBpore\n\n*Mycobacterium tuberculosis* genomic analysis from Nanopore sequencing data\n\n[![Python CI](https://github.com/mbhall88/tbpore/actions/workflows/ci.yaml/badge.svg)](https://github.com/mbhall88/tbpore/actions/workflows/ci.yaml)\n[![codecov](https://codecov.io/gh/mbhall88/tbpore/branch/main/graph/badge.svg)](https://codecov.io/gh/mbhall88/tbpore)\n[![PyPI](https://img.shields.io/pypi/v/tbpore)](https://pypi.org/project/tbpore/)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/tbpore)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n[TOC]: #\n\n# Table of Contents\n\n- [Synopsis](#synopsis)\n- [Citation](#citation)\n- [Installation](#installation)\n- [Configuring the decontamination database index](#configuring-the-decontamination-database-index)\n- [Performance](#performance)\n- [Usage](#usage)\n\n# Synopsis\n\n`tbpore` is a tool with two main goals.\nFirst is to process Nanopore Mycobacterium tuberculosis sequencing data to describe\nvariants with respect to the\ncanonical TB strain H37Rv and predict antibiotic resistance (command `tbpore process`).\nVariant description is done by decontaminating reads, calling variants with\n[bcftools](https://github.com/samtools/bcftools) and filtering variants.\nAntibiotic resistance is predicted\nwith [mykrobe](https://github.com/Mykrobe-tools/mykrobe).\nSecond, `tbpore` can be used to cluster TB samples based on their genotyping and a given\ndistance threshold (command\n`tbpore cluster`).\n\n## Citation\n\nTBpore is a slimmed-down version of\nthe [full pipeline](https://github.com/mbhall88/head_to_head_pipeline) used\nin our paper 👇\n\n\n> Hall, M. B. et al. Evaluation of Nanopore sequencing for Mycobacterium tuberculosis drug susceptibility testing and outbreak investigation: a genomic analysis. *The Lancet Microbe* 0, (2022) doi: [10.1016/S2666-5247(22)00301-9][doi].\n\n[doi]: https://doi.org/10.1016/S2666-5247(22)00301-9\n\n## Installation\n\n### conda\n\n[![Conda (channel only)](https://img.shields.io/conda/vn/bioconda/tbpore)](https://anaconda.org/bioconda/tbpore)\n[![bioconda version](https://anaconda.org/bioconda/tbpore/badges/platforms.svg)](https://anaconda.org/bioconda/tbpore)\n![Conda](https://img.shields.io/conda/dn/bioconda/tbpore)\n\nPrerequisite: [`conda`][conda] (and bioconda channel [correctly set up][channels])\n\n```shell\n$ conda install tbpore\n```\n\n### pip\n\n![PyPI](https://img.shields.io/pypi/v/tbpore)\n![PyPI - Python Version](https://img.shields.io/pypi/pyversions/tbpore)\n\nThe python components of `tbpore` are availble to install through [PyPI].\n\n```shell\npip install tbpore\n```\n\n**However**, you will need to install the following dependencies, which cannot be\ninstalled through PyPI.\n\n#### Dependencies\n\n* [`rasusa`](https://github.com/mbhall88/rasusa)\n* [`psdm`](https://github.com/mbhall88/psdm) version 0.1.x\n* [`samtools`](https://github.com/samtools/samtools) version 1.13\n* [`bcftools`](https://github.com/samtools/bcftools) version 1.13\n* [`mykrobe`](https://github.com/Mykrobe-tools/mykrobe) version 0.12.x\n* [`minimap2`](https://github.com/lh3/minimap2) version 2.22\n* [`seqkit`](https://bioinf.shenwei.me/seqkit/) version 2.x\n\nWe make no guarentees about the performance of `tbpore` with versions other than those\nspecified above. In particular, the `bcftools` version is very important. The latest\nversions of the other dependencies can likely be used.\n\n### Container\n\nDocker images are provided through biocontainers.\n\n#### `singularity`\n\nPrerequisite: [`singularity`][singularity]\n\n```shell\n$ URI="docker://quay.io/biocontainers/tbpore:<tag>"\n$ singularity exec "$URI" tbpore --help\n```\n\nsee [here][tags] for valid values for `<tag>`.\n\n#### `docker`\n\n[![Docker Repository on Quay](https://quay.io/repository/biocontainers/tbpore/status "Docker Repository on Quay")](https://quay.io/repository/biocontainers/tbpore)\n\nPrerequisite: [Docker]\n\n```shell\n$ docker pull quay.io/biocontainers/tbpore:<tag>\n$ docker run quay.io/biocontainers/tbpore:<tag> tbpore --help\n```\n\nsee [here][tags] for valid values for `<tag>`.\n\n## Configuring the decontamination database index\n\nAfter installing TBpore, you will need to download the decontamination database index.\n\n```\n$ tbpore download\n```\n\nBy default, this will download the index\nto `${HOME}/.tbpore/decontamination_db/remove_contam.map-ont.mmi`, as this is the\ndefault location `tbpore process` will search for.\n\nIf you prefer to download the index to another location, this can be done with\n\n```\n$ tbpore download -o other/location/db.mmi\n```\n\nKeep in mind, if you specify a non-default location, you will need to use the `--db`\noption when running `tbpore process`.\n\n## Performance\n\n### `tbpore process`\n\nBenchmarked on 151 TB ONT samples with 1 thread:\n\n* Runtime: `2103`s avg, `4048`s max (s = seconds);\n* RAM: `12.4`GB avg, `13.1`GB max (GB = Gigabytes);\n\n### `tbpore cluster`\n\nClustering 151 TB ONT samples:\n\n* Runtime: `286`s;\n* RAM: `<1`GB;\n\n## Usage\n\n### General usage\n\n```\nUsage: tbpore [OPTIONS] COMMAND [ARGS]...\n\nOptions:\n  -h, --help     Show this message and exit.\n  -V, --version  Show the version and exit.\n  -v, --verbose  Turns on debug-level logger. Option is mutually exclusive\n                 with quiet.\n  -q, --quiet    Turns off all logging except errors. Option is mutually\n                 exclusive with verbose.\n\nCommands:\n  cluster   Cluster consensus sequences\n  download  Download and validate the decontamination database\n  process   Single-sample TB genomic analysis from Nanopore sequencing data\n```\n\n### process\n\n```\nUsage: tbpore process [OPTIONS] [INPUTS]...\n\n  Single-sample TB genomic analysis from Nanopore sequencing data\n\n  INPUTS: Fastq file(s) and/or a directory containing fastq files. All files\n  will be joined into a single fastq file, so ensure they\'re all part of the\n  same sample/isolate.\n\nOptions:\n  -h, --help                      Show this message and exit.\n  -r, --recursive                 Recursively search INPUTS for fastq files\n  -S, --name TEXT                 Name of the sample. By default, will use the\n                                  first INPUT file with fastq extensions\n                                  removed\n  -A, --report_all_mykrobe_calls  Report all mykrobe calls (turn on flag -A,\n                                  --report_all_calls when calling mykrobe)\n  --db PATH                       Path to the decontaminaton database\n                                  [default: ${HOME}/.tbpore/decontamination_db/\n                                  remove_contam.map-ont.mmi]\n  -m, --metadata PATH             Path to the decontaminaton database metadata\n                                  file  [default: /Users/michaelhall/Projects/\n                                  tbpore/data/decontamination_db/remove_contam\n                                  .tsv.gz]\n  -o, --outdir DIRECTORY          Directory to place output files  [default:\n                                  .]\n  --tmp DIRECTORY                 Specify where to write all (tbpore)\n                                  temporary files. [default: <outdir>/.tbpore]\n  -t, --threads INTEGER           Number of threads to use in multithreaded\n                                  tools  [default: 1]\n  -d, --cleanup / -D, --no-cleanup\n                                  Remove all temporary files on *successful*\n                                  completion  [default: no-cleanup]\n  --cache DIRECTORY               Path to use for the cache  [default:\n                                  /Users/michaelhall/.cache]\n```\n\n### cluster\n\n```\nUsage: tbpore cluster [OPTIONS] [INPUTS]...\n\n  Cluster consensus sequences\n\n  Preferably input consensus sequences previously generated with tbpore\n  process.\n\n  INPUTS: Two or more consensus fasta sequences. Use glob patterns to input\n  several easily (e.g. output/sample_*/*.consensus.fa).\n\nOptions:\n  -h, --help                      Show this message and exit.\n  -T, --threshold INTEGER         Clustering threshold  [default: 6]\n  -o, --outdir DIRECTORY          Directory to place output files  [default:\n                                  .]\n  --tmp DIRECTORY                 Specify where to write all (tbpore)\n                                  temporary files. [default: <outdir>/.tbpore]\n  -t, --threads INTEGER           Number of threads to use in multithreaded\n                                  tools  [default: 1]\n  -d, --cleanup / -D, --no-cleanup\n                                  Remove all temporary files on *successful*\n                                  completion  [default: no-cleanup]\n  --cache DIRECTORY               Path to use for the cache  [default:\n                                  /Users/michaelhall/.cache]\n```\n\n### download\n\n```\nUsage: tbpore download [OPTIONS]\n\n  Download and validate the decontamination database\n\nOptions:\n  -h, --help         Show this message and exit.\n  -o, --output PATH  Download database to a specified filepath  [default: ${HOME}/\n                     .tbpore/decontamination_db/remove_contam.map-ont.mmi]\n  -f, --force        Force overwrite if the database already exists\n```\n\n[channels]: https://bioconda.github.io/#usage\n\n[conda]: https://docs.conda.io/projects/conda/en/latest/user-guide/install/\n\n[PyPI]: https://pypi.org/project/tbpore/\n\n[singularity]: https://sylabs.io/guides/3.6/user-guide/quick_start.html#quick-installation-steps\n\n[tags]: https://quay.io/repository/biocontainers/tbpore?tab=tags\n\n[Docker]: https://docs.docker.com/v17.12/install/\n',
    'author': 'Michael Hall',
    'author_email': 'michael@mbh.sh',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/mbhall88/tbpore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import functools
import gzip
import shutil
import sys
from pathlib import Path
from typing import Any, Dict, Tuple

import click
import yaml
from loguru import logger

from tbpore import (
    CACHE_DIR,
    CONFIG_FILE,
    DECONTAMINATION_DB_INDEX,
    DECONTAMINATION_DB_METADATA,
    EXTERNAL_SCRIPTS_DIR,
    TMP_NAME,
    H37RV_genome,
    H37RV_mask,
    __version__,
)
from tbpore.cli import Mutex
from tbpore.clustering import produce_clusters
from tbpore.external_tools import ExternalTool
from tbpore.utils import (
    concatenate_fastqs,
    decompress_file,
    download_file,
    fastq_prefix,
    find_fastq_files,
    parse_verbose_filter_params,
    validate_sha256,
)

log_fmt = (
    "[<green>{time:YYYY-MM-DD HH:mm:ss}</green>] <level>{level: <8}</level> | "
    "<level>{message}</level>"
)


def load_config_file() -> Dict[Any, Any]:
    with open(CONFIG_FILE, "r") as config_file_fh:
        return yaml.safe_load(config_file_fh)


def concatenate_inputs_into_infile(
    inputs: Tuple[Path, ...], infile: Path, recursive: bool, ctx: click.Context
) -> None:
    if len(inputs) == 1 and inputs[0].is_file():
        if not inputs[0].suffix == ".gz":
            with open(inputs[0], "rb") as f_in:
                with gzip.open(infile, "wb") as f_out:
                    shutil.copyfileobj(f_in, f_out)
        else:
            shutil.copy2(inputs[0], infile)
    else:
        logger.info("Searching for fastq files...")
        fq_files = set()
        for p in inputs:
            if p.is_file():
                fq_files.add(p)
            else:
                fq_files.update(find_fastq_files(p, recursive))

        # remove non-trivial duplicates (e.g. when comparing Path/str objects and paths referring to the same file but
        # represented differently)
        fq_files = set(map(lambda file: Path(file).resolve(), fq_files))

        no_fastq_files_found = len(fq_files) == 0
        if no_fastq_files_found:
            logger.error(
                "No fastq files found for the given inputs, please check your input files/dirs."
            )
            logger.error(
                "Tip: to search for fastq files recursively, please use the --recursive flag."
            )
            ctx.exit(2)

        logger.info(f"Found {len(fq_files)} fastq files. Joining them...")
        concatenate_fastqs(fq_files, infile)


def setup_logging(verbose: bool, quiet: bool) -> None:
    log_lvl = "INFO"
    if verbose:
        log_lvl = "DEBUG"
    elif quiet:
        log_lvl = "ERROR"
    logger.remove()
    logger.add(sys.stderr, level=log_lvl, format=log_fmt)


def setup_dirs(outdir: Path, tmp: Path, cache: Path) -> Tuple[Path, Path]:
    """
    Setups out, tmp, log and cache dirs, and return the paths to tmp and logdirs
    """
    outdir.mkdir(exist_ok=True, parents=True)
    if tmp is None:
        tmp = outdir / TMP_NAME
    tmp.mkdir(exist_ok=True, parents=True)
    logdir = outdir / "logs"
    cache.mkdir(parents=True, exist_ok=True)
    return tmp, logdir


def ensure_decontamination_db_is_available(
    path: Path,
    ctx: click.Context,
):
    if not path.exists():
        raise click.BadParameter(
            f"Decontamination DB index {path} does not exist, please run `tbpore "
            "download` to download and validate it before running tbpore",
            ctx=ctx,
            param_hint="--db",
        )


def common_opts(func):
    """Common CLI options for each subcommand. To add these options to a subcommand,
    use the decorator `@common_opts`
    Taken from https://stackoverflow.com/a/69082528/5299417
    """

    @click.option(
        "-o",
        "--outdir",
        help="Directory to place output files",
        default=".",
        show_default=True,
        type=click.Path(file_okay=False, writable=True, path_type=Path),
    )
    @click.option(
        "--tmp",
        help=(
            f"Specify where to write all (tbpore) temporary files. [default: "
            f"<outdir>/{TMP_NAME}]"
        ),
        type=click.Path(file_okay=False, writable=True, path_type=Path),
    )
    @click.option(
        "-t",
        "--threads",
        help="Number of threads to use in multithreaded tools",
        type=int,
        show_default=True,
        default=1,
    )
    @click.option(
        "--cleanup/--no-cleanup",
        "-d/-D",
        default=False,
        show_default=True,
        help="Remove all temporary files on *successful* completion",
    )
    @click.option(
        "--cache",
        type=click.Path(writable=True, file_okay=False, path_type=Path),
        help="Path to use for the cache",
        default=CACHE_DIR,
        show_default=True,
    )
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


@click.group()
@click.help_option("--help", "-h")
@click.version_option(__version__, "--version", "-V")
@click.option(
    "-v",
    "--verbose",
    help="Turns on debug-level logger. ",
    is_flag=True,
    cls=Mutex,
    not_required_if=["quiet"],
)
@click.option(
    "-q",
    "--quiet",
    help="Turns off all logging except errors. ",
    is_flag=True,
    cls=Mutex,
    not_required_if=["verbose"],
)
def main_cli(
    verbose: bool,
    quiet: bool,
):
    setup_logging(verbose, quiet)
    logger.info(f"Welcome to TBpore version {__version__}")


@main_cli.command()
@click.help_option("-h", "--help")
@click.option(
    "-r", "--recursive", help="Recursively search INPUTS for fastq files", is_flag=True
)
@click.option(
    "-S",
    "--name",
    help=(
        "Name of the sample. By default, will use the first INPUT file with fastq "
        "extensions removed"
    ),
)
@click.option(
    "-A",
    "--report_all_mykrobe_calls",
    is_flag=True,
    default=False,
    show_default=True,
    help="Report all mykrobe calls (turn on flag -A, --report_all_calls when calling mykrobe)",
)
@click.option(
    "--db",
    type=Path,
    help="Path to the decontaminaton database",
    default=DECONTAMINATION_DB_INDEX,
    show_default=True,
)
@click.option(
    "-m",
    "--metadata",
    type=click.Path(exists=True, path_type=Path),
    help="Path to the decontaminaton database metadata file",
    default=DECONTAMINATION_DB_METADATA,
    show_default=True,
)
@common_opts
@click.argument("inputs", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.pass_context
def process(
    ctx: click.Context,
    outdir: Path,
    inputs: Tuple[Path, ...],
    recursive: bool,
    tmp: Path,
    name: str,
    threads: int,
    report_all_mykrobe_calls: bool,
    cleanup: bool,
    db: Path,
    metadata: Path,
    cache: Path,
):
    """Single-sample TB genomic analysis from Nanopore sequencing data

    INPUTS: Fastq file(s) and/or a directory containing fastq files. All files will
    be joined into a single fastq file, so ensure they're all part of the same
    sample/isolate.
    """
    config = load_config_file()

    if not inputs:
        logger.error("No INPUT files given")
        ctx.exit(2)

    ensure_decontamination_db_is_available(db, ctx)

    if metadata == DECONTAMINATION_DB_METADATA and db != DECONTAMINATION_DB_INDEX:
        logger.info(
            "You have specified an alternate path to the decontamination database "
            "but are using the default metadata file. If you are using the tbpore-provided "
            "database, this is fine. Otherwise we recommend creating your own metadata "
            "file"
        )

    if metadata != DECONTAMINATION_DB_METADATA and db == DECONTAMINATION_DB_INDEX:
        logger.info(
            "You have specified a custom metadata file, but the default "
            "decontamination database. We trust you know what you're doing"
        )

    if not name:
        name = fastq_prefix(inputs[0])
        if not name:
            name = "input"
        logger.debug(f"No sample name found; using '{name}'")

    tmp, logdir = setup_dirs(outdir, tmp, cache)

    infile = tmp / f"{name}.fq.gz"
    concatenate_inputs_into_infile(inputs, infile, recursive, ctx)

    decontaminated_sam = str(tmp / f"{name}.decontaminated.sam")
    map_decontamination_db = ExternalTool(
        tool="minimap2",
        input=f"{db} {infile}",
        output=f"-o {decontaminated_sam}",
        params=f"{config['minimap2']['map_to_decom_DB']['params']} -t {threads}",
        logdir=logdir,
    )

    sorted_decontaminated_bam = str(tmp / f"{name}.decontaminated.sorted.bam")
    sort_decontaminated_sam = ExternalTool(
        tool="samtools",
        input=decontaminated_sam,
        output=f"-o {sorted_decontaminated_bam}",
        params=f"sort -@ {threads} {config['samtools']['sort_decom_DB']['params']}",
        logdir=logdir,
    )

    index_sorted_decontaminated_bam = ExternalTool(
        tool="samtools",
        input=sorted_decontaminated_bam,
        output="",
        params=f"index -@ {threads} {config['samtools']['index_decom_DB']['params']}",
        logdir=logdir,
    )

    filter_contamination_dir = tmp / f"{name}.decontaminated.filter"
    filter_contamination_dir.mkdir(parents=True, exist_ok=True)
    filter_contamination = ExternalTool(
        tool=sys.executable,
        input=f"-i {sorted_decontaminated_bam} -m {metadata}",
        output=f"-o {filter_contamination_dir}",
        params=f"{EXTERNAL_SCRIPTS_DIR / 'filter_contamination.py'} {config['filter_contamination']['params']}",
        logdir=logdir,
    )

    reads_to_keep = filter_contamination_dir / "keep.reads"
    decontaminated_nanopore_reads = tmp / f"{name}.decontaminated.fastq.gz"
    extract_decontaminated_nanopore_reads = ExternalTool(
        tool="seqkit",
        input=f"-f {reads_to_keep} {infile}",
        output=f"-o {decontaminated_nanopore_reads}",
        params=config["extract_decontaminated_nanopore_reads"]["params"],
        logdir=logdir,
    )

    sorted_fastq = tmp / f"{name}.sorted.fastq.gz"
    sort_decontaminated_reads = ExternalTool(
        tool="seqkit",
        input=str(decontaminated_nanopore_reads),
        output=f"-o {sorted_fastq}",
        params=config["sort_decontaminated_reads"]["params"],
        logdir=logdir,
    )

    subsampled_reads = f"{tmp}/{name}.subsampled.fastq.gz"
    rasusa = ExternalTool(
        tool="rasusa",
        input=f"-i {sorted_fastq}",
        output=f"-o {subsampled_reads}",
        params=config["rasusa"]["params"],
        logdir=logdir,
    )

    report_all_mykrobe_calls_param = "-A" if report_all_mykrobe_calls else ""
    mykrobe_output = f"{outdir}/{name}.mykrobe.json"
    mykrobe = ExternalTool(
        tool="mykrobe",
        input=f"-i {subsampled_reads}",
        output=f"-o {mykrobe_output}",
        params=f"predict {report_all_mykrobe_calls_param} --sample {name} -t {threads} --tmp {tmp} "
        f"--skeleton_dir {CACHE_DIR} {config['mykrobe']['predict']['params']}",
        logdir=logdir,
    )

    sam_file = f"{tmp}/{name}.subsampled.sam"
    minimap = ExternalTool(
        tool="minimap2",
        input=f"{H37RV_genome} {subsampled_reads}",
        output=f"-o {sam_file}",
        params=f"-t {threads} {config['minimap2']['map']['params']}",
        logdir=logdir,
    )

    sorted_sam_file = f"{tmp}/{name}.subsampled.sorted.sam"
    samtools_sort = ExternalTool(
        tool="samtools",
        input=sam_file,
        output=f"-o {sorted_sam_file}",
        params=f"sort -@ {threads} {config['samtools']['sort']['params']}",
        logdir=logdir,
    )

    pileup_file = f"{tmp}/{name}.pileup.vcf"
    bcftools_mpileup = ExternalTool(
        tool="bcftools",
        input=sorted_sam_file,
        output=f"-o {pileup_file}",
        params=f"mpileup -f {H37RV_genome} --threads {threads} {config['bcftools']['mpileup']['params']}",
        logdir=logdir,
    )

    snps_file = f"{tmp}/{name}.snps.vcf"
    bcftools_call = ExternalTool(
        tool="bcftools",
        input=pileup_file,
        output=f"-o {snps_file}",
        params=f"call --threads {threads} {config['bcftools']['call']['params']}",
        logdir=logdir,
    )

    filtered_snps_file = f"{outdir}/{name}.snps.filtered.bcf"
    filtering_options = " ".join(
        (
            config["filter"]["params"],
            parse_verbose_filter_params(config["filter"]["verbose_params"]),
        )
    )
    filter_vcf = ExternalTool(
        tool=sys.executable,
        input=f"-i {snps_file}",
        output=f"-o {filtered_snps_file}",
        params=f"{EXTERNAL_SCRIPTS_DIR / 'apply_filters.py'} {filtering_options}",
        logdir=logdir,
    )

    consensus_file = f"{outdir}/{name}.consensus.fa"
    generate_consensus = ExternalTool(
        tool=sys.executable,
        input=f"-i {filtered_snps_file} -f {H37RV_genome} -m {H37RV_mask}",
        output=f"-o {consensus_file}",
        params=f"{EXTERNAL_SCRIPTS_DIR / 'consensus.py'} --sample-id {name} {config['consensus']['params']}",
        logdir=logdir,
    )

    tools_to_run = (
        map_decontamination_db,
        sort_decontaminated_sam,
        index_sorted_decontaminated_bam,
        filter_contamination,
        extract_decontaminated_nanopore_reads,
        sort_decontaminated_reads,
        rasusa,
        mykrobe,
        minimap,
        samtools_sort,
        bcftools_mpileup,
        bcftools_call,
        filter_vcf,
        generate_consensus,
    )
    ExternalTool.run_tools(tools_to_run, ctx)

    if cleanup:
        logger.info("Cleaning up temporary files...")
        shutil.rmtree(tmp)

    logger.success("Done")


@main_cli.command()
@click.help_option("-h", "--help")
@click.option(
    "-T",
    "--threshold",
    help="Clustering threshold",
    type=int,
    show_default=True,
    default=6,
)
@common_opts
@click.argument("inputs", type=click.Path(exists=True, path_type=Path), nargs=-1)
@click.pass_context
def cluster(
    ctx: click.Context,
    threshold: int,
    outdir: Path,
    inputs: Tuple[Path, ...],
    tmp: Path,
    threads: int,
    cleanup: bool,
    cache: Path,
):
    """Cluster consensus sequences

    Preferably input consensus sequences previously generated with tbpore process.

    INPUTS: Two or more consensus fasta sequences. Use glob patterns to input several easily (e.g. output/sample_*/*.consensus.fa).
    """
    not_enough_input = len(inputs) < 2
    if not_enough_input:
        logger.error(
            "To cluster consensus sequences, please provide at least two input consensus sequences"
        )
        ctx.exit(2)

    config = load_config_file()

    tmp, logdir = setup_dirs(outdir, tmp, cache)

    infile = tmp / "all_sequences.fq.gz"
    concatenate_inputs_into_infile(inputs, infile, False, ctx)

    psdm_matrix = tmp / "psdm.matrix.csv"
    psdm = ExternalTool(
        tool="psdm",
        input=str(infile),
        output=f"-o {psdm_matrix}",
        params=f"{config['psdm']['params']} -t {threads}",
        logdir=logdir,
    )
    ExternalTool.run_tools((psdm,), ctx)

    logger.info("Producing clusters...")
    produce_clusters(psdm_matrix, threshold, outdir)
    logger.info("Done producing clusters")

    if cleanup:
        logger.info("Cleaning up temporary files...")
        shutil.rmtree(tmp)

    logger.success("Done")


@main_cli.command()
@click.help_option("-h", "--help")
@click.option(
    "-o",
    "--output",
    help="Download database to a specified filepath",
    type=Path,
    default=DECONTAMINATION_DB_INDEX,
    show_default=True,
)
@click.option(
    "-f", "--force", help="Force overwrite if the database already exists", is_flag=True
)
@click.pass_context
def download(ctx: click.Context, output: Path, force: bool):
    """Download and validate the decontamination database"""
    if output.exists() and not force:
        logger.error(
            f"Output path {output} already exists and --force was not used. Exiting to "
            "avoid overwriting existing file"
        )
        ctx.exit(2)

    if not output.parent.exists():
        output.parent.mkdir(parents=True)

    db_config = load_config_file().get("decontamination_db")
    if db_config is None:
        logger.error("Missing expected key 'decontamination_db' in config file")
        ctx.exit(2)

    url = db_config.get("url")
    if url is None:
        logger.error("Missing URL for decontamination database")
        ctx.exit(2)

    expected_compressed_hash = db_config.get("compressed_sha256")
    if expected_compressed_hash is None:
        logger.error("Missing compressed sha256 hash for decontamination database")
        ctx.exit(2)

    expected_decompressed_hash = db_config.get("decompressed_sha256")
    if expected_decompressed_hash is None:
        logger.error("Missing decompressed sha256 hash for decontamination database")
        ctx.exit(2)

    logger.info(
        f"Downloading decontamination database to {output}. This may take a while..."
    )
    compressed_output = Path(f"{output}.gz")
    download_file(url, compressed_output)

    logger.info("Validating the decontamination database...")
    if not validate_sha256(compressed_output, expected_compressed_hash):
        logger.error(
            "The sha256 hash for the compressed decontamination database did not match "
            "the expected hash"
        )
        ctx.exit(2)
    logger.success("Downloaded and validated the decontamination database")

    logger.info("Decompressing the decontamination database...")
    decompress_file(compressed_output, output, remove_compressed=True)

    logger.info("Validating the decompressed decontamination database...")
    if not validate_sha256(output, expected_decompressed_hash):
        logger.error(
            "The sha256 hash for the decompressed decontamination database did not match the expected hash"
        )
        ctx.exit(2)
    logger.success("Decompressed and validated the decontamination database")

    logger.success("Done!")


def main():
    main_cli()


if __name__ == "__main__":
    main()
